#include "stock.h"
#include "threads.h"

/* ds-routines.c - contains routines for basic data structures.. 
 * like operations on arrays, stacks, linked lists etc... */

req_list_t *head_req_list;
req_list_t *tail_req_list;

void buy_pr_remove (cid_t cid, plid_t plid, int posn) {
   /* Since this code removes 1 quote, 
    * player[0] has to quote here.. to fill the void..
    * cause nature abhors a vacuum..
    */
   
   register int j;
   company_t *c = &company[cid];
   stock_company_t *t = c->buy[posn];
   
   t->plid = 0;
   t->qty = 0;
   t->pr = 0;
   
   for (j=posn; j < last_quot; ++j)
     c->buy[j] = c->buy[j+1];
   
   c->buy[last_quot] = t;
   
   return;
}
   
void buy_pr_insert (cid_t cid, plid_t plid, pr_t pr, qty_t qty) {
   register int i, j;
   company_t *c = &company[cid];
   stock_company_t *t = c->buy[last_quot];
	    
   /* This following piece of code can be written more efficiently,
    * reducing the 2 nested loops to 1... later...
    */
   
   for (i=0; i < max_simult_quot; ++i) {
      if (pr > c->buy[i]->pr) {
	 
	 for (j=last_quot; j > i; --j)
	   c->buy[j] = c->buy[j-1];

	 c->buy[i] = t;
	 
	 t->plid = plid;
	 t->pr = pr;
	 t->qty = qty;
	 
         break;
      }
   }   
}
   
void sell_pr_remove (cid_t cid, plid_t plid, int posn) {
   register int j;
   company_t *c= &company[cid];
   stock_company_t *t = c->sell[posn];
   
   t->plid = 0;
   t->qty = 0;
   t->pr = -1;
   
   for (j=posn; j < last_quot; ++j)
     c->sell[j] = c->sell[j+1];
   
   c->sell[last_quot] = t;

   return;
}

void sell_pr_insert (cid_t cid, plid_t plid, pr_t pr, qty_t qty) {
   register int i,j;
   company_t *c = &company[cid];
   stock_company_t *t = c->sell[last_quot];
   
   for (i=0; i < max_simult_quot; ++i) {
      if (pr < c->sell[i]->pr) {
	 
	 for (j=last_quot; j > i; --j )
	   c->sell[j] = c->sell[j-1];

	 c->sell[i] = t;
	 
	 t->plid = plid;
	 t->pr = pr;
	 t->qty = qty;
	 
         break;
      }
   }
}


/* Hash table routines, insertion and searching...
 * each element of the hash table is the head of a linked list.
 * NOTE : linked list insertion happens at the head
 */

stock_player_t *holding_insert (plid_t plid, cid_t cid, qty_t qty) {
   register int posn;
   stock_player_t *t;
   player_t *p = &player[plid];
   
   posn = cid % hashsize;
   t = (stock_player_t *) xmalloc (sizeof(stock_player_t));
   
   t->cid = cid;
   t->qty = qty;
   
   t->next = p->holding[posn];
   p->holding[posn] = t;
   return t;
}

stock_player_t *holding_search (plid_t plid, cid_t cid) {
   register int posn;
   stock_player_t *t;
   
   posn = cid % hashsize;
   t = player[plid].holding[posn];
   
   while (t != NULL) {
      if (t->cid == cid) 
	return t;
      t = t->next;
   }
   return NULL;
}


int addto_req_list (int client, plid_t plid, cid_t cid,
		   void (*call_fn)
		   (int client, plid_t plid, cid_t cid, pr_t pr, qty_t qty)) {
   pr_t pr;
   qty_t qty;
   
   sr_pr (pr);
   sr_qty (qty);
   
   if (tail_req_list->next == head_req_list)
     return 0;
   
   pthread_mutex_lock (&req_list_mut);  /* CR for req_list */

   tail_req_list->pr = pr;
   tail_req_list->qty = qty;
   tail_req_list->plid = plid;
   tail_req_list->cid = cid;
   tail_req_list->call_fn = call_fn;

   pthread_mutex_unlock (&req_list_mut);

#ifdef LOG
   fprintf (logfile, "pr %d  qty %d\n", tail_req_list->pr, tail_req_list->qty);
#endif
   
   tail_req_list = tail_req_list->next;
   pthread_mutex_unlock (&req_list_mut);
   
   return 1;
}

void addto_result_list (plid_t plid, req_t req,
			cid_t cid, qty_t qty, pr_t pr) {
   
   player_t *p = &player[plid];
   result_list_t *t = p->tail;
   pthread_mutex_t *result_mut = &p->result_mut;
   
   pthread_mutex_lock (result_mut); /* CR to protect cond, list */
   
   /* If we have no more room, then we ask for more...
    * the extra malloc calls increase the cost per transaction, but
    * then who said that the world is perfect..
    */
   if (t->next == p->head) 
     grow_result_list (t, resize);
   
   t->req = req;
   t->cid = cid;
   t->qty = qty;
   t->pr = pr;
   
   p->tail = t->next;

   pthread_cond_signal (&p->result_cond); /* wakeup sender thread */
   pthread_mutex_unlock (result_mut);
   
   return;
}

void setup_req_list () {
   register int i;
   req_list_t *t;
   
   t = (req_list_t *) xmalloc (sizeof(req_list_t));
   tail_req_list = head_req_list = t;
   
   for (i=0; i < 100; ++i) {
      t->next = (req_list_t *) xmalloc (sizeof(req_list_t));
      t = t->next;
   }
   t->next = head_req_list;
   
}

void setup_result_list (plid_t plid) {
   register int i;
   result_list_t *t;
   player_t *p = &player[plid];
   
   t = (result_list_t *) xmalloc (sizeof(result_list_t));
   p->tail = p->head = t;
   
   for (i=0; i < resize; ++i) {
      t->next = (result_list_t *) xmalloc (sizeof(result_list_t));
      t = t->next;
   }
   
   t->next = p->head;   
}

void grow_result_list (result_list_t *node, int n) {
   register int i;
   result_list_t *t1, *t2;
   
   t1 = (result_list_t *) xmalloc (sizeof(result_list_t));
   t2 = node->next;
   node->next = t1;
   
   for (i=0; i < n; ++i) {
      t1->next = (result_list_t *) xmalloc (sizeof(result_list_t));
      t1 = t1->next;
   }
   t1->next = t2;
   return;
}

