#include "stock.h"
#include "threads.h"

#include <netinet/in.h>
#include <sys/socket.h>
#include <errno.h>

static int establish (int portnum);
static void init_server (int server);
static void init_public ();
static void init_admin_threads ();
void thread_stuff ();
void signal_stuff ();

int server2;
pthread_t trader_th;
pthread_t cleaner_th;
pthread_t life_th;

int main (int argc, char *argv[]) {
   int server;
   
   init_defaults ();
   parse_options (argc, argv);
   signal_stuff ();
   thread_stuff ();
   init_companies ();
   init_public ();
   init_admin_threads ();
   setup_req_list ();
   
   if ((server = establish (portnum)) < 0) {
      printf ("Couldn't open socket for listening\n");
      exit (2);
   }
     
   if ((server2 = establish (++portnum)) < 0) {
      printf ("Couldn't open socket for listening\n");
      exit(2);
   }
       
   init_server (server);
   /* ooops, we're not supposed to be here, now what did Micosoft do again? */
   /* I don't want to go anywhere today. */
   return -1;
}

/* code to establish a socket */

static int establish (int portnum) {
   int server;
   int reuse_addr = 1;
   struct sockaddr_in sa;
   
   memset(&sa, 0, sizeof(struct sockaddr_in));
   sa.sin_family = AF_INET;
   sa.sin_port= htons(portnum);
   sa.sin_addr.s_addr = htonl(INADDR_ANY);
   
   if ((server = socket(AF_INET, SOCK_STREAM, 0)) < 0)  
     return(-1);
   
   setsockopt (server, SOL_SOCKET, SO_REUSEADDR, &reuse_addr, 
	       sizeof (reuse_addr));
   
   if (bind(server, &sa, sizeof(struct sockaddr_in)) < 0) {
      perror ("bind");
      close(server);
      return(-2);
   }
   
   if (listen (server, 5) < 0) {
      perror ("listen");
      close (server);
      return (-3);
   }
   
   return(server);
}

static void init_public () {
   register int i;
   qty_t qty = 10000;
   player_t *p = &player[0];

   p->holding = (stock_player_t **) xmalloc (hashsize * sizeof(stock_player_t *));
   
   for (i=0; i <= max_companies; ++i)
     holding_insert (0, i, qty);
   
   p->name = "Public Holding\0";
   p->client = p->client2 = 0;
   p->cash = -1;
   return;
}

void init_admin_threads () {
      
   /* Start thread for the trading, i.e. clearing the req queue */
   
   if (pthread_create (&trader_th, &attr_detached, trader, NULL) != 0) {
#ifdef DEBUG
      printf ("Trouble creating thread\n");
#endif
      exit (5);
   }

   /* start the cleaner thread */
   
   if (pthread_create (&cleaner_th, &attr_detached, cleaner, NULL) < 0) {
#ifdef DEBUG
      printf ("Trouble creating thread\n");
#endif
      exit (5);
   }
   
   /* Starting the GAME OF LIFE Thread */
   
#ifdef LIFE
   
   if (pthread_create (&life_th, &attr_detached, life, NULL) < 0) {
#ifdef DEBUG
      printf ("Trouble creating thread\n");
#endif
      exit (5);
   }

#endif  /* LIFE */
   
}

static void init_server (int server) {
   
#ifdef DEBUG
   printf ("Server Ready to accept connections.\n");
#endif
   
   while (1) {
      int client;
      thread_arg_t *thread_arg;
      pthread_t *sync_th;
      struct sockaddr_in client_addr;
      int client_addrlen;
      
      /* Ideally, we should check no. of players and sleep
       * if no new logins are allowed...
       * This feature will appear soon.. watch this place..
       */
      
      if ((client = accept(server, &client_addr, &client_addrlen)) < 0) {
	 if (errno == EINTR) continue;  
#ifdef DEBUG
	 perror("accept");
#endif
	 exit(2);
      }      
      
#ifdef LOG
      fprintf (logfile, "New Connection from %s\n", 
	       inet_ntoa (client_addr.sin_addr));
#endif
	
      thread_arg = (thread_arg_t *) xmalloc (sizeof(thread_arg_t));
      sync_th = (pthread_t *) xmalloc (sizeof(pthread_t));            
      thread_arg->client = client;
      thread_arg->th = sync_th;
      
      if (pthread_create (sync_th, &attr_detached, 
			  serve_requests_sync, (void *) thread_arg) < 0) {
#ifdef DEBUG
	 printf ("Error creating thread\n");
#endif
	 close (client);
	 free (thread_arg);
      }
   }  /* while (1) */
}  /* init_server () */
